Package: GCPtools
Title: Tools for working with gcloud and gsutil
Version: 1.3.2
Date: 2026-05-22
Authors@R: c(
    person("Marcel", "Ramos", , "marcel.ramos@sph.cuny.edu",
        c("aut", "cre"), c(ORCID = "0000-0002-3242-0582")
    ),
    person("Nitesh", "Turaga", role = "aut"),
    person("Martin", "Morgan", , , "aut", c(ORCID = "0000-0002-5874-8148"))
    )
Description: Lower-level functionality to interface with Google Cloud Platform
    tools. 'gcloud' and 'gsutil' are both supported. The functionality provided
    centers around utilities for the AnVIL platform.
biocViews: Software, Infrastructure, ThirdPartyClient, DataImport
SystemRequirements: gsutil, gcloud
Depends: R (>= 4.6.0)
Imports: AnVILBase, BiocBaseUtils, dplyr, httr, rlang, tibble, tidyr,
        utils
Suggests: BiocStyle, knitr, rmarkdown, testthat (>= 3.0.0)
License: Artistic-2.0
URL: https://github.com/Bioconductor/GCPtools
BugReports: https://github.com/Bioconductor/GCPtools/issues
Encoding: UTF-8
VignetteBuilder: knitr
Roxygen: list(markdown = TRUE)
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/GCPtools
git_branch: devel
git_last_commit: ee812c9
git_last_commit_date: 2026-05-22
Repository: Bioconductor 3.24
Date/Publication: 2026-05-24
NeedsCompilation: no
Packaged: 2026-05-24 21:58:08 UTC; biocbuild
Author: Marcel Ramos [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3242-0582>),
  Nitesh Turaga [aut],
  Martin Morgan [aut] (ORCID: <https://orcid.org/0000-0002-5874-8148>)
Maintainer: Marcel Ramos <marcel.ramos@sph.cuny.edu>

# Generated by roxygen2: do not edit by hand

S3method(print,gcloud_sdk_result)
export(gcloud_access_token)
export(gcloud_account)
export(gcloud_cmd)
export(gcloud_exists)
export(gcloud_help)
export(gcloud_project)
export(gcloud_storage)
export(gcloud_storage_buckets)
export(gsutil_cat)
export(gsutil_cp)
export(gsutil_exists)
export(gsutil_help)
export(gsutil_is_uri)
export(gsutil_ls)
export(gsutil_pipe)
export(gsutil_requesterpays)
export(gsutil_requesterpays_flag)
export(gsutil_rm)
export(gsutil_rsync)
export(gsutil_sh_quote)
export(gsutil_stat)
importFrom(AnVILBase,avstop_for_status)
importFrom(BiocBaseUtils,isCharacter)
importFrom(BiocBaseUtils,isScalarCharacter)
importFrom(BiocBaseUtils,isScalarLogical)
importFrom(BiocBaseUtils,isZeroOneCharacter)
importFrom(httr,POST)
importFrom(httr,content)
importFrom(httr,content_type)
importFrom(rlang,.data)
importFrom(tidyr,pivot_wider)
importFrom(utils,tail)

.gcloud_do <-
    function(...)
{
    .gcloud_sdk_do("gcloud", c(...))
}

#' @importFrom httr POST content_type content
#' @importFrom AnVILBase avstop_for_status
.gcloud_access_token_new <-
    function(app_default, now)
{
    ## obtain the access token
    token <- .gcloud_do("auth", app_default, "print-access-token")

    ## There is only one token per service account, so requesting a
    ## token may return one with expiration less than 60 minutes. So
    ## check the actual expiry time of the token.
    ##
    ## Calculating expiry from before the call (`now` argument) is
    ## conservative -- we underestimate the time by the latency
    ## involved in the POST and result parsing.
    response <- POST(
        "https://www.googleapis.com/oauth2/v1/tokeninfo",
        encode = "form",
        body = list(access_token = token)
    )
    avstop_for_status(response, ".gcloud_access_token_expires")
    expires <- now + content(response)$expires_in

    list(token = token, expires = expires)
}

#' @rdname gcloud
#'
#' @name gcloud
#'
#' @title gcloud command line utility interface
#'
#' @description These functions invoke the `gcloud` command line utility. See
#'   [gsutil] for details on how `gcloud` is located.
NULL


#' @name gcloud_access_token
#'
#' @title Obtain an access token for a service account
#'
#' @param service `character(1)` The name of the service, e.g. "terra" for which
#'   to obtain an access token for.
#'
#' @description `gcloud_access_token()` generates a token for the given service
#'   account. The token is cached for the duration of its validity. The token is
#'   refreshed when it expires. The token is obtained using the `gcloud` command
#'   line utility for the given `gcloud_account()`. The function is mainly used
#'   internally by API service functions, e.g., `AnVIL::Terra()`
#'
#' @return `gcloud_access_token()` returns a simple token string to be used with
#'   the given service.
#'
#' @examplesIf gcloud_exists()
#' gcloud_access_token("rawls") |> invisible()
#' @export
gcloud_access_token <- local({
    tokens <- new.env(parent = emptyenv())
    function(service) {
        app_default <-
            if (identical(Sys.getenv("USER"), "jupyter-user"))
                "application-default"

        key <- paste0(service, ":", app_default, ":", gcloud_account())
        now <- Sys.time()
        if (is.null(tokens[[key]])) {
            tokens[[key]] <- .gcloud_access_token_new(app_default, now)
        } else {
            expires_in <- tokens[[key]]$expires - now
            if (expires_in < 1L) {
                ## allow a nearly expired token to fully expire
                if (expires_in > 0L)
                    Sys.sleep(expires_in)
                tokens[[key]] <- .gcloud_access_token_new(app_default, now)
            }
        }

        tokens[[key]]$token
    }
})

gcloud_runs <- function() {
    tryCatch({
        .gcloud_do("version")[1L] |>
            startsWith("Google Cloud SDK")
    }, error = function(...) FALSE)
}

#' @rdname gcloud
#'
#' @description `gcloud_exists()` tests whether the `gcloud()` command
#'     can be found on this system. After finding the binary location,
#'     it runs `gcloud version` to identify potentially misconfigured
#'     installations. See 'Details' section of `gsutil` for where the
#'     application is searched.
#'
#' @return `gcloud_exists()` returns `TRUE` when the `gcloud`
#'     application can be found, FALSE otherwise.
#'
#' @examples
#' gcloud_exists()
#' @export
gcloud_exists <-
    function()
{
    result <- tryCatch({
        .gcloud_sdk_find_binary("gcloud")
    }, error = function(...) "")
    nzchar(result) && gcloud_runs()
}

#' @importFrom utils tail
.gcloud_get_value_check <-
    function(result, function_name)
{
    value <- tail(result, 1L)
    if (identical(value, "(unset)")) {
        message <- paste0(
            "'", function_name, "()' returned '(unset)'; this may indicate ",
            "that the gcloud active configuration is incorrect. Try ",
            "`gcloud auth application-default login` at the command line"
        )
        warning(paste(strwrap(message), collapse = "\n"))
    }
    value
}

#' @rdname gcloud
#'
#' @description `gcloud_account()`: report the current gcloud account
#'     via `gcloud config get-value account`.
#'
#' @param account character(1) Google account (e.g., `user@gmail.com`)
#'     to use for authentication.
#'
#' @return `gcloud_account()` returns a `character(1)` vector
#'     containing the active gcloud account, typically a gmail email
#'     address.
#'
#' @importFrom BiocBaseUtils isScalarCharacter
#'
#' @examplesIf gcloud_exists()
#' gcloud_account()
#' @export
gcloud_account <- function(account = NULL) {
    stopifnot(is.null(account) || isScalarCharacter(account))

    if (!is.null(account))
        .gcloud_do("config", "set", "account", account)
    result <- .gcloud_do("config", "get-value", "account")
    .gcloud_get_value_check(result, "gcloud_account")
}

#' @rdname gcloud
#'
#' @description `gcloud_project()`: report the current gcloud project
#'     via `gcloud config get-value project`.
#'
#' @param project character(1) billing project name.
#'
#' @return `gcloud_project()` returns a `character(1)` vector
#'     containing the active gcloud project.
#'
#' @export
gcloud_project <- function(project = NULL) {
    stopifnot(
        is.null(project) || isScalarCharacter(project)
    )

    if (!is.null(project))
        .gcloud_do("config", "set", "project", project)
    result <- .gcloud_do("config", "get-value", "project")
    ## returns two lines when `CLOUDSDK_ACTIVE_CONFIG_NAME=`
    ## environment variable is set
    .gcloud_get_value_check(result, "gcloud_project")
}

#' @rdname gcloud
#'
#' @description `gcloud_help()`: queries `gcloud` for help for a
#'     command or sub-comand via `gcloud help ...`.
#'
#' @param ... Additional arguments appended to `gcloud` commands.
#'
#' @return `gcloud_help()` returns an unquoted `character()` vector
#'     representing the text of the help manual page returned by
#'     `gcloud help ...`.
#'
#' @export
gcloud_help <- function(...)
    .gcloud_sdk_result(.gcloud_do("help", ...))

#' @rdname gcloud
#'
#' @description `gcloud_cmd()` allows arbitrary `gcloud` command
#'     execution via `gcloud ...`. Use pre-defined functions in
#'     preference to this.
#'
#' @param cmd `character(1)` representing a command used to evaluate
#'     `gcloud cmd ...`.
#'
#' @return `gcloud_cmd()` returns a `character()` vector representing
#'     the text of the output of `gcloud cmd ...`
#'
#' @export
gcloud_cmd <- function(cmd, ...)
    .gcloud_do(cmd, ...)

#' @rdname gcloud
#'
#' @description `gcloud_storage()` allows arbitrary `gcloud storage` command
#'   execution via `gcloud storage ...`. Typically used for bucket management
#'   commands such as `rm` and `cp`.
#'
#' @export
gcloud_storage <- function(cmd, ...)
    .gcloud_do("storage", cmd, ...)

#' @rdname gcloud
#'
#' @description `gcloud_storage_buckets()` provides an interface to the
#'  `gcloud storage buckets` command. This command can be used to create a new
#'   bucket via `gcloud storage buckets create ...`.
#'
#' @param bucket_cmd `character(1)` representing a buckets command typically
#'   used to create a new bucket. It can also be used to
#'   `add-iam-policy-binding` or `remove-iam-policy-binding` to a bucket.
#'
#' @param bucket `character(1)` representing a unique bucket name to be created
#'   or modified.
#'
#' @importFrom BiocBaseUtils isScalarCharacter
#' @export
gcloud_storage_buckets <- function(bucket_cmd = "create", bucket, ...) {
    stopifnot(
        isScalarCharacter(bucket_cmd), isScalarCharacter(bucket)
    )
    gcloud_storage("buckets", bucket_cmd, bucket, ...)
}

context("gcloud_sdk")

test_that(".gcloud_sdk_find_binary() works", {
    sdk <- tempfile()
    dir.create(file.path(sdk, "bin", "gsutil"), recursive = TRUE)
    sdk <- normalizePath(sdk)
    expected <- normalizePath(file.path(sdk, "bin", "gsutil"))

    withr::with_envvar(c(GCLOUD_SDK_PATH=sdk), {
        object <- .gcloud_sdk_find_binary("gsutil")
        expect_identical(object, expected)
    })
})

test_that("gcloud_project() returns correctly when config unset", {
    skip_if(!gcloud_exists())
    withr::with_envvar(c(CLOUDSDK_ACTIVE_CONFIG_NAME="__UNDEFINED__"), {
        expect_warning(object <- gcloud_project())
        expect_identical(object, "(unset)")
    })
})

.gcloud_storage_do <- function(...)
{
    .gcloud_sdk_do("gcloud", c("storage", ...))
}

#' @rdname gsutil
#'
#' @name gsutil
#'
#' @title gsutil command line utility interface
#'
#' @description These functions invoke the `gsutil` command line
#'     utility. See the "Details:" section if you have gsutil
#'     installed but the package cannot find it.
#'
#' @details The `gsutil` system command is required.  The search for
#'     `gsutil` starts with environment variable `GCLOUD_SDK_PATH`
#'     providing a path to a directory containing a `bin` directory
#'     containingin `gsutil`, `gcloud`, etc. The path variable is
#'     searched for first as an `option()` and then system
#'     variable. If no option or global variable is found,
#'     `Sys.which()` is tried. If that fails, `gsutil` is searched for
#'     on defined paths. On Windows, the search tries to find
#'     `Google\\Cloud SDK\\google-cloud-sdk\\bin\\gsutil.cmd` in the
#'     `LOCAL APP DATA`, `Program Files`, and `Program Files (x86)`
#'     directories.  On linux / macOS, the search continues with
#'     `~/google-cloud-sdk`.
#'
#' @examples
#' ## use a truly public dataset for testing
#' src <- paste0(
#'   "gs://gcp-public-data-landsat/",
#'   "LC08/01/001/002/LC08_L1GT_001002_20160902_20170321_01_T2/",
#'   "LC08_L1GT_001002_20160902_20170321_01_T2_MTL.txt"
#' )
NULL

## evaluate the gsutil command and arguments in `args`
.gsutil_do <-
    function(args)
{
    .gcloud_sdk_do("gsutil", args)
}

#' @rdname gsutil
#'
#' @description `gsutil_is_uri()`: check if the source is a valid
#'     Google Storage URI.
#'
#' @param source `character(1)`, (`character()` for `gsutil_requesterpays()`,
#'   `gsutil_ls()`, `gsutil_exists()`, `gsutil_cp()`): paths to a Google Storage
#'   Bucket, possibly with wild-cards for file-level pattern matching.
#'
#' @importFrom BiocBaseUtils isCharacter
#'
#' @export
gsutil_is_uri <-
    function(source)
{
    isCharacter(source) & grepl("^gs://[^/]+", source)
}

#' @rdname gsutil
#'
#' @description `gsutil_sh_quote()`: quote a character vector for
#'   use in a shell command. This is useful to ensure that file
#'   names with spaces or other special characters are handled
#'   correctly.
#'
#' @export
gsutil_sh_quote <-
    function(source)
{
    ## Expand local paths with ~ or . or .. to full path names.
    ## Needed because we also use shQuote() (to allow for spaces in
    ## file names), and shQuote() would otherwise use paths with ~ or
    ## . in the current working directory.
    is_local <- !gsutil_is_uri(source)
    source[is_local] <- normalizePath(source[is_local])
    shQuote(source)
}

#' @rdname gsutil
#'
#' @description `gsutil_requesterpays()`: does the google bucket
#'     require that the requester pay for access?
#'
#' @return `gsutil_requesterpays()`: named `logical()` vector TRUE
#'     when requester-pays is enabled.
#'
#' @examplesIf gcloud_exists()
#' gsutil_requesterpays(src) # FALSE -- no cost download
#'
#' @export
gsutil_requesterpays <-
    function(source)
{
    stopifnot(all(gsutil_is_uri(source)))
    project <- gcloud_project()
    buckets <- regmatches(source, regexpr("^gs://[^/]+", source))
    is_enabled <- FALSE
    for (bucket in buckets) {
        args <- c("-u", project, "requesterpays", "get", bucket)
        result <- .gsutil_do(args)
        is_enabled <- any(endsWith(result, "Enabled"))
        if (is_enabled)
            break
    }
    is_enabled
}

#' @rdname gsutil
#'
#' @description `gsutil_requesterpays_flag()`: return the `-u <project>` flag
#'   for the `gsutil` command if the source requires that the requester pays
#'   for access.
#'
#' @return `gsutil_requesterpays_flag()`: `character()` vector of
#'   `"-u <gcloud_project>"` if the source requires that the requester pays
#'   for access, or `NULL` if otherwise.
#'
#' @export
gsutil_requesterpays_flag <-
    function(source)
{
    source <- source[gsutil_is_uri(source)]
    tryCatch({
        if (length(source) && gsutil_requesterpays(source))
            c("-u", gcloud_project())
        else
            NULL
    }, error = function(e) {
        ## this was originally written to return NULL without a
        ## warning, but I'm not sure whether we cannot just stop()?
        warning(
            "'gsutil_requesterpays()' returned an error:",
            "\n  ", conditionMessage(e),
            call. = FALSE
        )
        NULL
    })
}

.gsutil_exists_1 <-
    function(source, gsutil)
{
    args <- c(
        gsutil_requesterpays_flag(source),
        "ls",
        shQuote(source)
    )
    value <- withCallingHandlers({
        system2(gsutil, args, stdout = TRUE, stderr = TRUE, wait=TRUE)
    }, warning = function(w) {
        invokeRestart("muffleWarning")
    })
    is.null(attr(value, "status"))
}

#' @rdname gsutil
#'
#' @description `gsutil_exists()`: check if the bucket or object
#'     exists.
#'
#' @return `gsutil_exists()`: logical(1) TRUE if bucket or object exists.
#'
#' @export
gsutil_exists <-
    function(source)
{
    stopifnot(
        is.character(source), !anyNA(source),
        gsutil_is_uri(source)
    )

    gsutil <- .gcloud_sdk_find_binary("gsutil")
    stopifnot(file.exists(gsutil))      # bad environment variables

    source <- stats::setNames(nm = source)
    vapply(source, .gsutil_exists_1, logical(1), gsutil)
}

#' @rdname gsutil
#'
#' @description `gsutil_stat()`: print, as a side effect, the status
#'     of a bucket, directory, or file.
#'
#' @return `gsutil_stat()`: `tibble()` summarizing status of each
#'     bucket member.
#'
#' @importFrom tidyr pivot_wider
#' @importFrom rlang .data
#'
#' @examplesIf gcloud_exists()
#' gsutil_exists(src)
#' gsutil_stat(src)
#'
#' @export
gsutil_stat <-
    function(source)
{
    stopifnot(gsutil_is_uri(source))

    args <- c(gsutil_requesterpays_flag(source), "stat", shQuote(source))
    result <- .gsutil_do(args)

    ## omit nested 'metadata', for convenience
    is_metadata <- grepl("^( {4}Metadata:| {8})", result)
    result <- result[!is_metadata]

    ## form into tibble with rows for each bucket & key / value pair
    is_path <- startsWith(result, "gs://")
    group <- cumsum(is_path)
    n <- tabulate(group) - 1L
    re <- "^ +([^:]+): +(.*)"
    tz_format <- "%a, %d %b %Y %H:%M:%S"
    tbl <- tibble::tibble(
        path = rep(sub(":$", "", result[is_path]), n),
        key = sub(re, "\\1", result[!is_path]),
        value = sub(re, "\\2", result[!is_path])
    )

    ## reshape to one row per bucket
    tbl |>
        tidyr::pivot_wider(
            id_cols = .data$path, names_from = "key", values_from = "value"
        ) |>
        dplyr::mutate(
            `Creation time` =
                as.POSIXct(
                    .data$`Creation time`, tz = "GMT", format = tz_format
                ),
            `Update time` =
                as.POSIXct(.data$`Update time`, tz = "GMT", format = tz_format)
        )
}

#' @name gsutil
#'
#' @description `gsutil_rsync()`: synchronize a source and a
#'     destination. If the destination is on the local file system, it
#'     must be a directory or not yet exist (in which case a directory
#'     will be created).
#'
#' @param destination `character(1)` google cloud bucket or local file system
#'   destination path.
#'
#' @param ... additional arguments passed as-is to the `gsutil` subcommand.
#'
#' @param exclude `character(1)` a python regular expression of bucket
#'     paths to exclude from synchronization. E.g.,
#'     `'.*(\\.png|\\.txt)$"` excludes '.png' and .txt' files.
#'
#' @param dry `logical(1)`, when `TRUE` (default), return the
#'     consequences of the operation without actually performing the
#'     operation.
#'
#' @param delete `logical(1)`, when `TRUE`, remove files in
#'     `destination` that are not in `source`. Exercise caution when
#'     you use this option: it's possible to delete large amounts of
#'     data accidentally if, for example, you erroneously reverse
#'     source and destination.
#'
#' @param recursive `logical(1)`; perform operation recursively from `source`?.
#'   Default: `FALSE`.
#'
#' @param parallel `logical(1)`, perform parallel multi-threaded /
#'   multi-processing (default is `TRUE`).
#'
#' @details
#'
#' `gsutil_rsync()': To make `"gs://mybucket/data"` match the contents
#' of the local directory `"data"` you could do:
#'
#' `gsutil_rsync("data", "gs://mybucket/data", delete = TRUE)`
#'
#' To make the local directory "data" the same as the contents of
#' gs://mybucket/data:
#'
#' `gsutil_rsync("gs://mybucket/data", "data", delete = TRUE)`
#'
#' If `destination` is a local path and does not exist, it will be
#' created.
#'
#' @return `gsutil_rsync()`: exit status of `gsutil_rsync()`, invisibly.
#'
#' @importFrom BiocBaseUtils isScalarLogical
#'
#' @export
gsutil_rsync <-
    function(source, destination, ..., exclude = NULL, dry = TRUE,
        delete = FALSE, recursive = FALSE, parallel = TRUE)
{
    stopifnot(
        isScalarCharacter(source), isScalarCharacter(destination),
        gsutil_is_uri(source) || gsutil_is_uri(destination),
        is.null(exclude) || isScalarCharacter(exclude),
        isScalarLogical(dry),
        isScalarLogical(delete),
        isScalarLogical(recursive),
        isScalarLogical(parallel)
    )
    ## if destination is not a google cloud repo, and does not exist
    if (!dry && !gsutil_is_uri(destination) && !dir.exists(destination))
        if (!dir.create(destination))
            stop("'gsutil_rsync()' failed to create '", destination, "'")

    ## rsync operation
    args <- c(
        gsutil_requesterpays_flag(source),
        ##  -m option, to perform parallel (multi-threaded/multi-processing)
        if (parallel) "-m",
        "rsync",
        if (length(exclude)) c("-x", shQuote(exclude)),
        if (dry) "-n",
        if (delete) "-d",
        if (recursive) "-r",
        ...,
        gsutil_sh_quote(source),
        gsutil_sh_quote(destination)
    )
    result <- .gsutil_do(args)
    .gcloud_sdk_result(result)
}

#' @rdname gsutil
#'
#' @description `gsutil_cat()`: concatenate bucket objects to standard output
#'
#' @param header `logical(1)` when `TRUE` annotate each
#'
#' @param range (optional) `integer(2)` vector used to form a range
#'     from-to of bytes to concatenate. `NA` values signify
#'     concatenation from the start (first position) or to the end
#'     (second position) of the file.
#'
#' @return `gsutil_cat()` returns the content as a character vector.
#'
#' @importFrom BiocBaseUtils isScalarCharacter
#'
#' @export
gsutil_cat <-
    function(source, ..., header = FALSE, range = integer())
{
    stopifnot(
        isScalarCharacter(source),
        isScalarLogical(header),
        is.numeric(range),
        all(range[!is.na(range)] >= 0),
        all(diff(range[!is.na(range)]) > 0L),
        length(range) == 0L || length(range) == 2L
    )

    if (length(range)) {
        range[is.na(range)] <- ""
        range <- paste(range, collapse="-")
    }

    args <- c(
        gsutil_requesterpays_flag(source),
        "cat",
        if (header) "-h",
        if (length(range)) c("-r", range),
        shQuote(source)
    )

    .gsutil_do(args)
}

#' @rdname gsutil
#'
#' @description `gsutil_help()`: print 'man' page for the `gsutil`
#'     command or subcommand. Note that only commandes documented on this
#'     R help page are supported.
#'
#' @param cmd `character()` (optional) command name, e.g.,
#'     `"ls"` for help.
#'
#' @return `gsutil_help()`: `character()` help text for subcommand `cmd`.
#'
#' @importFrom BiocBaseUtils isZeroOneCharacter
#'
#' @examplesIf gcloud_exists()
#' gsutil_help("ls")
#'
#' @export
gsutil_help <-
    function(cmd = character(0))
{
    stopifnot(isZeroOneCharacter(cmd))
    result <- .gsutil_do(c("help", cmd))
    .gcloud_sdk_result(result)
}

##
## higher-level implementations
##

#' @rdname gsutil
#'
#' @description `gsutil_pipe()`: create a pipe to read from or write
#'     to a Google bucket object.
#'
#' @param open `character(1)` either `"r"` (read) or `"w"` (write)
#'     from the bucket.
#'
#' @return `gsutil_pipe()` an unopened R `pipe()`; the mode is
#'     \emph{not} specified, and the pipe must be used in the
#'     appropriate context (e.g., a pipe created with `open = "r"` for
#'     input as `readLines()`)
#'
#' @examplesIf gcloud_exists()
#' lines <- readLines(gsutil_pipe(src))
#' length(lines)
#' head(lines)
#'
#' @export
gsutil_pipe <-
    function(source, open = "r", ...)
{
    stopifnot(
        isScalarCharacter(source),
        isScalarCharacter(open)
    )

    is_read <- identical(substr(open, 1, 1), "r")
    args <- c(
        if (is_read) gsutil_requesterpays_flag(source),
        "cp",
        ...,
        if (is_read) c(source, "-") else c("-", source)
    )
    args <- vapply(args, shQuote, character(1L))

    bin <- .gcloud_sdk_find_binary("gsutil")
    stopifnot(file.exists(bin))

    cmd <- paste(c(shQuote(bin), args), collapse = " ")
    pipe(cmd, open)
}

#' @rdname gsutil
#'
#' @description `gsutil_ls()`: List contents of a google cloud bucket
#'   or, if `source` is missing, all Cloud Storage buckets under your
#'   default project ID
#'
#' @return `gsutil_ls()`: `character()` listing of `source` content.
#'
#' @export
gsutil_ls <-
    function(source = character(), ..., recursive = FALSE)
{
    stopifnot(
        gsutil_is_uri(source),
        isScalarLogical(recursive)
    )

    args <- c(
        gsutil_requesterpays_flag(source),
        "ls",
        if (recursive) "-r",
        ...,
        shQuote(source)
    )
    result <- .gsutil_do(args)
    result[nzchar(result) & !endsWith(result, ":")]
}

#' @rdname gsutil
#'
#' @description `gsutil_cp()`: copy contents of `source` to
#'   `destination`. At least one of `source` or `destination` must
#'   be Google cloud bucket; `source` can be a character vector with
#'   length greater than 1. Use `gsutil_help("cp")` for `gsutil` help.
#'
#' @return `gsutil_cp()`: exit status of `gsutil_cp()`, invisibly.
#'
#' @examplesIf gcloud_exists()
#'   gsutil_cp(src, tempdir())
#'   ## gsutil_*() commands work with spaces in the source or destination
#'   destination <- file.path(tempdir(), "foo bar")
#'   gsutil_cp(src, destination)
#'   file.exists(destination)
#' @export
gsutil_cp <-
    function(source, destination, ..., recursive = FALSE, parallel = TRUE)
{
    location <- c(source, destination)
    location_is_uri <- gsutil_is_uri(location)
    stopifnot(
        isCharacter(source), isScalarCharacter(destination),
        any(location_is_uri),
        isScalarLogical(recursive), isScalarLogical(parallel)
    )

    args <- c(
        gsutil_requesterpays_flag(location),
        if (parallel) "-m", ## Makes the operations faster
        "cp", ## cp command
        if (recursive) "-r",
        ...,
        gsutil_sh_quote(source),
        gsutil_sh_quote(destination)
    )
    result <- .gsutil_do(args)
    .gcloud_sdk_result(result)
}

#' @rdname gsutil
#'
#' @description `gsutil_rm()`: remove contents of a google cloud
#'     bucket.
#'
#' @param force `logical(1)`: continue silently despite errors when
#'     removing multiple objects. Default: `FALSE`.
#'
#' @return `gsutil_rm()`: exit status of `gsutil_rm()`, invisibly.
#'
#' @export
gsutil_rm <-
    function(source, ..., force = FALSE, recursive = FALSE, parallel = TRUE)
{
    stopifnot(
        gsutil_is_uri(source),
        isScalarLogical(force),
        isScalarLogical(recursive),
        isScalarLogical(parallel)
    )

    ## remove
    args <- c(
        gsutil_requesterpays_flag(source),
        if (parallel) "-m",
        "rm",
        if (force) "-f",
        if (recursive) "-r",
        ...,
        shQuote(source)
    )
    result <- .gsutil_do(args)
    .gcloud_sdk_result(result)
}

context("gsutil")

test_that("'gsutil_is_uri()' works", {
    expect_identical(gsutil_is_uri(character()), logical(0))
    expect_identical(gsutil_is_uri("gs://bucket"), TRUE)
    expect_identical(gsutil_is_uri(c("gs://bucket", "/path")), c(TRUE, FALSE))
    expect_identical(gsutil_is_uri(1), FALSE)
})

test_that("gsutil_sh_quote()' works", {
    home_dir <- "~"
    bucket <- "gs://foo"
    expect_identical(gsutil_sh_quote(home_dir), shQuote(path.expand(home_dir)))
    expect_identical(gsutil_sh_quote(bucket), shQuote(bucket))

    non_existent_dir <- tempfile(tmpdir="~")
    expect_warning(gsutil_sh_quote(non_existent_dir))
    suppressWarnings({
        expect_identical(
            gsutil_sh_quote(non_existent_dir),
            shQuote(normalizePath(non_existent_dir))
        )
    })

    all <- c(home_dir, bucket, non_existent_dir)
    expect_warning(gsutil_sh_quote(all))
    suppressWarnings({
        expect_identical(
            gsutil_sh_quote(all),
            shQuote(c(
                normalizePath(home_dir),
                bucket,
                normalizePath(non_existent_dir)
            ))
        )
    })
})

##
## gcloud_sdk_result constructor and methods
##
.gcloud_sdk_result <-
    function(x)
{
    if (!is.null(x))
        class(x) <- "gcloud_sdk_result"
    x
}

#' @export
print.gcloud_sdk_result <-
    function(x, ...)
{
    if (is.null(x))
        return()
    cat(noquote(x), sep="\n")
}


## option or environment variable or NULL, allowing for default
## `unset` value
.gcloud_sdk_getenv <-
    function(option, unset = NA)
{
    value <- getOption(option, unset)
    if (!is.na(value))
        return(value)

    value <- Sys.getenv(option, unset = unset)
    if (!is.na(value))
        return(value)

    return(NULL)
}

## Get gsutil binary on user's machine for windows/linux/mac
.gcloud_sdk_find_binary <-
    function(binary_name)
{
    user_path <- .gcloud_sdk_getenv("GCLOUD_SDK_PATH")
    if (!is.null(user_path))
        return(normalizePath(file.path(user_path, "bin", binary_name)))

    bin <- Sys.which(binary_name)
    if (nzchar(bin))
        return(normalizePath(bin))

    ## Discover binary automatically if user doesn't give path
    if (.Platform$OS.type == "windows") {
        appdata <- normalizePath(Sys.getenv("localappdata"), winslash = "/")
        sdk_path <- file.path("Google", "Cloud SDK", "google-cloud-sdk", "bin")
        binary_cmd <- paste(binary_name, "cmd", sep = ".")
        bin_paths <- c(
            file.path(appdata, sdk_path, binary_cmd),
            file.path(Sys.getenv("ProgramFiles"), sdk_path, binary_cmd),
            file.path(Sys.getenv("ProgramFiles(x86)"), sdk_path, binary_cmd)
        )
    } else {
        bin_paths <- file.path("~", "google-cloud-sdk", "bin", binary_name)
    }

    ## Return appropriate path for 'gsutil'
    for (path in bin_paths)
        if (file.exists(path))
            return(normalizePath(path))

    stop(
        "failed to find '", binary_name, "' binary; ",
        "set option or environment variable 'GCLOUD_SDK_PATH'?",
        call. = FALSE
    )
}

#' @importFrom BiocBaseUtils isScalarCharacter isCharacter
.gcloud_sdk_do <-
    function(command, args)
{
    stopifnot(
        isScalarCharacter(command),
        isCharacter(args, na.ok = FALSE)
    )
    bin <- .gcloud_sdk_find_binary(command)
    stopifnot(file.exists(bin))

    value <- withCallingHandlers({
        tryCatch({
            system2(bin, args, stdout = TRUE, stderr = TRUE, wait=TRUE)
        }, error = function(err) {
            msg <- paste0(
                "'", command, " ", paste(args, collapse = " "), "' failed:\n",
                "  ", conditionMessage(err)
            )
            stop(msg, call. = FALSE)
        })
    }, warning = function(warn) {
        invokeRestart("muffleWarning")
    })

    if (!is.null(attr(value, "status"))) {
        errmsg <- attr(value, "errmsg")
        if (!is.null(errmsg) && nzchar(errmsg))
            errmsg <- paste0("\n  condition message: ", errmsg)
        msg <- paste0(
            "'", command, " ", paste(args, collapse = " "), "' failed:",
            "\n  ", paste(as.vector(value), collapse = "\n    "),
            "\n  exit status: ", attr(value, "status"),
            errmsg
        )
        stop(msg, call. = FALSE)
    }

    value
}

---
title: "GCPtools"
author: "AnVIL Team"
date: "`r format(Sys.Date(), '%B %d, %Y')`"
output: 
  BiocStyle::html_document:
    toc: true
    toc_depth: 2
    toc_float: true
    number_sections: true
package: GCPtools
vignette: >
  %\VignetteIndexEntry{GCPtools Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# GCPtools

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
has_gcloud <- GCPtools::gcloud_exists()
options(width = 75)
```

GCPtools is a package that provides an interface to Google Cloud Platform (GCP)
tools, primarily the `gcloud` and `gsutil` utilities. It allows users to
interact with GCP services such as Google Cloud Storage (GCS) and Google Compute
Engine (GCE) directly from R.

# Installation and Load

You can install the latest version of `GCPtools` from GitHub using the
following:

```
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("Bioconductor/GCPtools")
```

After installation, you can load the package using:

```{r load-gcp-tools,include=TRUE,results="hide",message=FALSE,warning=FALSE}
library(GCPtools)
```

# `gcloud` Command Line Utility

## Check for `gcloud` SDK

The `gcloud_exists()` function is used to check if the `gcloud` command line
utility is available on your system.

```{r gcloud-exists}
gcloud_exists()
```

## `gcloud` Help

The package provides a function to display the help for the `gcloud` command
line utility. This can be useful to understand the available commands and
options. They can be used in conjunction with `gcloud_cmd()` to run
specific commands.


```{r gcloud_help,eval=has_gcloud}
gcloud_help() |>
    head(10L)
```

## Access Token

The package provides functions to interact with GCP services. For example, you
can use `gcloud_access_token()` to obtain an access token for a specific
service. This is currently in use by `AnVIL` and `AnVILGCP` R packages.

```{r gcloud-access-token,eval=FALSE}
gcloud_access_token("terra")
```

## Command Execution

You can execute `gcloud` commands using the `gcloud_cmd()` function. This
function allows you to run any `gcloud` command directly from R, making it easy
to interact with GCP services without leaving the R environment.

```{r gcloud-cmd,eval=has_gcloud}
gcloud_cmd("version")
```

## List of available functions

The package provides a variety of functions to interact with GCP services.

* `gcloud_exists()`: tests whether the `gcloud()` command can be found.
* `gcloud_account()`: report the current gcloud account via ⁠gcloud config
    get-value account⁠.
* `gcloud_project()`: report the current gcloud project via
    `⁠gcloud config get-value project`⁠.
* `gcloud_help()`: queries gcloud for help for a command or sub-comand via
    `⁠gcloud help ...`.
* `gcloud_cmd()`: `gcloud` command execution via `⁠gcloud ...`. Use pre-defined
    functions in preference to this.
* `gcloud_storage()`: `⁠gcloud` storage⁠ command execution via
    `⁠gcloud storage ...`⁠. Typically used for bucket management commands such as
    `rm` and `cp`.
* `gcloud_storage_buckets()`: interface to the `⁠gcloud storage buckets`⁠
    command. It can create a new bucket via `⁠gcloud storage buckets create ...⁠`.

# `gsutil` Command Line Utility

## Check for `gsutil` resources

The `gsutil_exists()` function checks if a bucket or object exists in Google
Cloud Storage (GCS).

```{r gsutil-exists,eval=FALSE}
gsutil_exists("gs://your-bucket-name")
```

## `gsutil` Help

The `gsutil_help()` function displays the help for the `gsutil` command line
utility. This can be useful to understand the available commands and options
and how to use them effectively.

```{r gsutil_help,eval=has_gcloud}
gsutil_help()
```

The package also includes functions to manage Google Cloud
Storage buckets and objects, such as `gsutil_ls()` to list objects in a bucket
and `gsutil_cp()` to copy files to and from GCS.

## List of available functions

The package provides a variety of functions to interact with GCS.

* `gsutil_requesterpays()`: does the Google bucket require that the requester
    pay for access?
* `gsutil_requesterpays_flag()`: return the `-u <gcloud_project>` flag
    if the bucket requires that the requester pay for access.
* `gsutil_exists()`: check if the bucket or object exists.
* `gsutil_stat()`: print, as a side effect, the status of a bucket, directory,
    or file.
* `gsutil_rsync()`: synchronize a source and a destination. 
* `gsutil_cat()`: concatenate bucket objects to standard output
* `gsutil_help()`: print 'man' page for the `gsutil` command or subcommand.
* `gsutil_pipe()`: create a pipe to read from or write to a Google bucket
    object.

# Disclaimer

This package is not an official Google product, nor is it affiliated with or
endorsed by Google LLC. It is developed and maintained by the AnVIL team to
provide a convenient interface to GCP tools for R users on the AnVIL platform.

Only the above set of `gcloud` and `gsutil` commands are supported.

# SessionInfo

```{r session-info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gcloud.R
\name{gcloud}
\alias{gcloud}
\alias{gcloud_exists}
\alias{gcloud_account}
\alias{gcloud_project}
\alias{gcloud_help}
\alias{gcloud_cmd}
\alias{gcloud_storage}
\alias{gcloud_storage_buckets}
\title{gcloud command line utility interface}
\usage{
gcloud_exists()

gcloud_account(account = NULL)

gcloud_project(project = NULL)

gcloud_help(...)

gcloud_cmd(cmd, ...)

gcloud_storage(cmd, ...)

gcloud_storage_buckets(bucket_cmd = "create", bucket, ...)
}
\arguments{
\item{account}{character(1) Google account (e.g., \code{user@gmail.com})
to use for authentication.}

\item{project}{character(1) billing project name.}

\item{...}{Additional arguments appended to \code{gcloud} commands.}

\item{cmd}{\code{character(1)} representing a command used to evaluate
\verb{gcloud cmd ...}.}

\item{bucket_cmd}{\code{character(1)} representing a buckets command typically
used to create a new bucket. It can also be used to
\code{add-iam-policy-binding} or \code{remove-iam-policy-binding} to a bucket.}

\item{bucket}{\code{character(1)} representing a unique bucket name to be created
or modified.}
}
\value{
\code{gcloud_exists()} returns \code{TRUE} when the \code{gcloud}
application can be found, FALSE otherwise.

\code{gcloud_account()} returns a \code{character(1)} vector
containing the active gcloud account, typically a gmail email
address.

\code{gcloud_project()} returns a \code{character(1)} vector
containing the active gcloud project.

\code{gcloud_help()} returns an unquoted \code{character()} vector
representing the text of the help manual page returned by
\verb{gcloud help ...}.

\code{gcloud_cmd()} returns a \code{character()} vector representing
the text of the output of \verb{gcloud cmd ...}
}
\description{
These functions invoke the \code{gcloud} command line utility. See
\link{gsutil} for details on how \code{gcloud} is located.

\code{gcloud_exists()} tests whether the \code{gcloud()} command
can be found on this system. After finding the binary location,
it runs \verb{gcloud version} to identify potentially misconfigured
installations. See 'Details' section of \code{gsutil} for where the
application is searched.

\code{gcloud_account()}: report the current gcloud account
via \verb{gcloud config get-value account}.

\code{gcloud_project()}: report the current gcloud project
via \verb{gcloud config get-value project}.

\code{gcloud_help()}: queries \code{gcloud} for help for a
command or sub-comand via \verb{gcloud help ...}.

\code{gcloud_cmd()} allows arbitrary \code{gcloud} command
execution via \verb{gcloud ...}. Use pre-defined functions in
preference to this.

\code{gcloud_storage()} allows arbitrary \verb{gcloud storage} command
execution via \verb{gcloud storage ...}. Typically used for bucket management
commands such as \code{rm} and \code{cp}.

\code{gcloud_storage_buckets()} provides an interface to the
\verb{gcloud storage buckets} command. This command can be used to create a new
bucket via \verb{gcloud storage buckets create ...}.
}
\examples{
gcloud_exists()
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
gcloud_account()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gsutil.R
\name{gsutil}
\alias{gsutil}
\alias{gsutil_is_uri}
\alias{gsutil_sh_quote}
\alias{gsutil_requesterpays}
\alias{gsutil_requesterpays_flag}
\alias{gsutil_exists}
\alias{gsutil_stat}
\alias{gsutil_rsync}
\alias{gsutil_cat}
\alias{gsutil_help}
\alias{gsutil_pipe}
\alias{gsutil_ls}
\alias{gsutil_cp}
\alias{gsutil_rm}
\title{gsutil command line utility interface}
\usage{
gsutil_is_uri(source)

gsutil_sh_quote(source)

gsutil_requesterpays(source)

gsutil_requesterpays_flag(source)

gsutil_exists(source)

gsutil_stat(source)

gsutil_rsync(
  source,
  destination,
  ...,
  exclude = NULL,
  dry = TRUE,
  delete = FALSE,
  recursive = FALSE,
  parallel = TRUE
)

gsutil_cat(source, ..., header = FALSE, range = integer())

gsutil_help(cmd = character(0))

gsutil_pipe(source, open = "r", ...)

gsutil_ls(source = character(), ..., recursive = FALSE)

gsutil_cp(source, destination, ..., recursive = FALSE, parallel = TRUE)

gsutil_rm(source, ..., force = FALSE, recursive = FALSE, parallel = TRUE)
}
\arguments{
\item{source}{\code{character(1)}, (\code{character()} for \code{gsutil_requesterpays()},
\code{gsutil_ls()}, \code{gsutil_exists()}, \code{gsutil_cp()}): paths to a Google Storage
Bucket, possibly with wild-cards for file-level pattern matching.}

\item{destination}{\code{character(1)} google cloud bucket or local file system
destination path.}

\item{...}{additional arguments passed as-is to the \code{gsutil} subcommand.}

\item{exclude}{\code{character(1)} a python regular expression of bucket
paths to exclude from synchronization. E.g.,
\verb{'.*(\\\\.png|\\\\.txt)$"} excludes '.png' and .txt' files.}

\item{dry}{\code{logical(1)}, when \code{TRUE} (default), return the
consequences of the operation without actually performing the
operation.}

\item{delete}{\code{logical(1)}, when \code{TRUE}, remove files in
\code{destination} that are not in \code{source}. Exercise caution when
you use this option: it's possible to delete large amounts of
data accidentally if, for example, you erroneously reverse
source and destination.}

\item{recursive}{\code{logical(1)}; perform operation recursively from \code{source}?.
Default: \code{FALSE}.}

\item{parallel}{\code{logical(1)}, perform parallel multi-threaded /
multi-processing (default is \code{TRUE}).}

\item{header}{\code{logical(1)} when \code{TRUE} annotate each}

\item{range}{(optional) \code{integer(2)} vector used to form a range
from-to of bytes to concatenate. \code{NA} values signify
concatenation from the start (first position) or to the end
(second position) of the file.}

\item{cmd}{\code{character()} (optional) command name, e.g.,
\code{"ls"} for help.}

\item{open}{\code{character(1)} either \code{"r"} (read) or \code{"w"} (write)
from the bucket.}

\item{force}{\code{logical(1)}: continue silently despite errors when
removing multiple objects. Default: \code{FALSE}.}
}
\value{
\code{gsutil_requesterpays()}: named \code{logical()} vector TRUE
when requester-pays is enabled.

\code{gsutil_requesterpays_flag()}: \code{character()} vector of
\code{"-u <gcloud_project>"} if the source requires that the requester pays
for access, or \code{NULL} if otherwise.

\code{gsutil_exists()}: logical(1) TRUE if bucket or object exists.

\code{gsutil_stat()}: \code{tibble()} summarizing status of each
bucket member.

\code{gsutil_rsync()}: exit status of \code{gsutil_rsync()}, invisibly.

\code{gsutil_cat()} returns the content as a character vector.

\code{gsutil_help()}: \code{character()} help text for subcommand \code{cmd}.

\code{gsutil_pipe()} an unopened R \code{pipe()}; the mode is
\emph{not} specified, and the pipe must be used in the
appropriate context (e.g., a pipe created with \code{open = "r"} for
input as \code{readLines()})

\code{gsutil_ls()}: \code{character()} listing of \code{source} content.

\code{gsutil_cp()}: exit status of \code{gsutil_cp()}, invisibly.

\code{gsutil_rm()}: exit status of \code{gsutil_rm()}, invisibly.
}
\description{
These functions invoke the \code{gsutil} command line
utility. See the "Details:" section if you have gsutil
installed but the package cannot find it.

\code{gsutil_is_uri()}: check if the source is a valid
Google Storage URI.

\code{gsutil_sh_quote()}: quote a character vector for
use in a shell command. This is useful to ensure that file
names with spaces or other special characters are handled
correctly.

\code{gsutil_requesterpays()}: does the google bucket
require that the requester pay for access?

\code{gsutil_requesterpays_flag()}: return the \verb{-u <project>} flag
for the \code{gsutil} command if the source requires that the requester pays
for access.

\code{gsutil_exists()}: check if the bucket or object
exists.

\code{gsutil_stat()}: print, as a side effect, the status
of a bucket, directory, or file.

\code{gsutil_rsync()}: synchronize a source and a
destination. If the destination is on the local file system, it
must be a directory or not yet exist (in which case a directory
will be created).

\code{gsutil_cat()}: concatenate bucket objects to standard output

\code{gsutil_help()}: print 'man' page for the \code{gsutil}
command or subcommand. Note that only commandes documented on this
R help page are supported.

\code{gsutil_pipe()}: create a pipe to read from or write
to a Google bucket object.

\code{gsutil_ls()}: List contents of a google cloud bucket
or, if \code{source} is missing, all Cloud Storage buckets under your
default project ID

\code{gsutil_cp()}: copy contents of \code{source} to
\code{destination}. At least one of \code{source} or \code{destination} must
be Google cloud bucket; \code{source} can be a character vector with
length greater than 1. Use \code{gsutil_help("cp")} for \code{gsutil} help.

\code{gsutil_rm()}: remove contents of a google cloud
bucket.
}
\details{
The \code{gsutil} system command is required.  The search for
\code{gsutil} starts with environment variable \code{GCLOUD_SDK_PATH}
providing a path to a directory containing a \code{bin} directory
containingin \code{gsutil}, \code{gcloud}, etc. The path variable is
searched for first as an \code{option()} and then system
variable. If no option or global variable is found,
\code{Sys.which()} is tried. If that fails, \code{gsutil} is searched for
on defined paths. On Windows, the search tries to find
\verb{Google\\\\Cloud SDK\\\\google-cloud-sdk\\\\bin\\\\gsutil.cmd} in the
\verb{LOCAL APP DATA}, \verb{Program Files}, and \verb{Program Files (x86)}
directories.  On linux / macOS, the search continues with
\verb{~/google-cloud-sdk}.

\verb{gsutil_rsync()': To make }"gs://mybucket/data"\verb{match the contents of the local directory}"data"` you could do:

\code{gsutil_rsync("data", "gs://mybucket/data", delete = TRUE)}

To make the local directory "data" the same as the contents of
gs://mybucket/data:

\code{gsutil_rsync("gs://mybucket/data", "data", delete = TRUE)}

If \code{destination} is a local path and does not exist, it will be
created.
}
\examples{
## use a truly public dataset for testing
src <- paste0(
  "gs://gcp-public-data-landsat/",
  "LC08/01/001/002/LC08_L1GT_001002_20160902_20170321_01_T2/",
  "LC08_L1GT_001002_20160902_20170321_01_T2_MTL.txt"
)
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
gsutil_requesterpays(src) # FALSE -- no cost download
\dontshow{\}) # examplesIf}
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
gsutil_exists(src)
gsutil_stat(src)
\dontshow{\}) # examplesIf}
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
gsutil_help("ls")
\dontshow{\}) # examplesIf}
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
lines <- readLines(gsutil_pipe(src))
length(lines)
head(lines)
\dontshow{\}) # examplesIf}
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
  gsutil_cp(src, tempdir())
  ## gsutil_*() commands work with spaces in the source or destination
  destination <- file.path(tempdir(), "foo bar")
  gsutil_cp(src, destination)
  file.exists(destination)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gcloud.R
\name{gcloud_access_token}
\alias{gcloud_access_token}
\title{Obtain an access token for a service account}
\usage{
gcloud_access_token(service)
}
\arguments{
\item{service}{\code{character(1)} The name of the service, e.g. "terra" for which
to obtain an access token for.}
}
\value{
\code{gcloud_access_token()} returns a simple token string to be used with
the given service.
}
\description{
\code{gcloud_access_token()} generates a token for the given service
account. The token is cached for the duration of its validity. The token is
refreshed when it expires. The token is obtained using the \code{gcloud} command
line utility for the given \code{gcloud_account()}. The function is mainly used
internally by API service functions, e.g., \code{AnVIL::Terra()}
}
\examples{
\dontshow{if (gcloud_exists()) withAutoprint(\{ # examplesIf}
gcloud_access_token("rawls") |> invisible()
\dontshow{\}) # examplesIf}
}
